# -*- coding: utf-8 -*-
import shutil
import os
from .Decorators import validate
from typing import Union
from pathlib import Path


@validate(str, list)
def write_to_file(path: str, lines: Union[list[str], list[bytes]], write_bytes: bool = False) -> None:
    """clear and then write data to file

    Args:
        path (str): path of file
        lines (list[str]): data to write
    """

    try:
        if write_bytes:
            with open(path, "wb") as f:
                f.writelines(lines)
        else:
            with open(path, "w", encoding="utf-8") as f:
                f.writelines(lines)
    except Exception as e:
        if isinstance(e, TypeError):
            raise Exception(
                "'lines' contains a 'bytes' object.\nTo use with bytes use: write_bytes = True ")
        raise e


@validate(str)
def path_exists(path: str) -> bool:
    """checks whether a path exists

    Args:
        path (str): path to check

    Returns:
        bool: result of check
    """
    return os.path.exists(path)


@validate(str)
def file_exists(path: str) -> bool:
    """checks whether a file exists at specified path

    Args:
        path (str): path to check

    Returns:
        bool: will return true iff the path exists and it is a path to a file
    """
    return path_exists(path) and is_file(path)


@validate(str)
def directory_exists(path: str) -> bool:
    """checks whether a directory exists at specified path

    Args:
        path (str): path to check

    Returns:
        bool: will return true iff the path exists and it is a path to a directory
    """
    return path_exists(path) and is_directory(path)


@validate(str)
def delete_file(path: str) -> None:
    """deletes a file if it exists

    Args:
        path (str): path of file
    """
    if file_exists(path):
        os.remove(path)


@validate(str, bool)
def read_file(path: str, read_bytes: bool = False) -> list[str]:
    """read all lines from a file

    Args:
        path (str): the path to the file

    Returns:
        list[str]: a list of all the lines in the file
    """
    try:
        if read_bytes:
            with open(path, "rb") as f:
                return f.readlines()
        else:
            with open(path, "r", encoding="mbcs") as f:
                return f.readlines()
    except Exception as e:
        if isinstance(e, UnicodeDecodeError):
            raise Exception(
                f"Can't read byte in file.\nTo use with bytes use: read_bytes = True ")
        raise e


@validate(str)
def is_file(path: str) -> bool:
    """return whether a path represents a file

    Args:
        path (str): path to check
    """
    return os.path.isfile(path)


@validate(str)
def is_directory(path: str) -> bool:
    """return whether a path represents a directory

    Args:
        path (str): path to check
    """
    return os.path.isdir(path)


@validate(str)
def get_files(path: str) -> list[str]:
    """return a list of names of all files inside specified directory

    Args:
        path (str): directory

    Returns:
        list[str]: all files
    """
    files_and_directories = get_files_and_directories(path)
    return list(
        filter(lambda name: is_file(f"{path}\\{name}"), files_and_directories))


@validate(str)
def get_files_and_directories(path: str) -> list[str]:
    """get a list of all files and directories in specified path

    Args:
        path (str): path to check

    Returns:
        list[str]: all files and directories
    """
    return os.listdir(path)


@validate(str)
def get_directories(path: str) -> list[str]:
    """get all directories in specified path

    Args:
        path (str): path to check

    Returns:
        list[str]: all directories
    """
    files_and_directories = get_files_and_directories(path)
    return list(
        filter(lambda name: is_directory(f"{path}\\{name}"), files_and_directories))


@ validate(str)
def delete_directory(path: str) -> None:
    """delete a directory and all its contents

    Args:
        path (str): _description_
    """
    if is_directory(path):
        clear_directory(path)
        os.rmdir(path)


@validate(str)
def clear_directory(path: str) -> None:
    for file in get_files(path):
        delete_file(f"{path}\\{file}")
    for dir in get_directories(path):
        delete_directory(f"{path}\\{dir}")


@validate(str)
def create_directory(path: str) -> None:
    """create a directory at the specified path if it doesn't already exists

    Args:
        path (str): the path to create a directory at
    """
    if not directory_exists(path):
        os.makedirs(path)


@validate(str, str)
def get_file_type_from_directory(path: str, file_type: str) -> list[str]:
    return list(
        filter(
            lambda name: Path(f"{path}\\{name}").suffix == file_type,
            get_files(path)
        )
    )


@validate(str, str)
def get_file_type_from_directory_recursively(path: str, file_type: str):
    res = []
    for dir in get_directories(path):
        res.extend(f"{dir}\\{v}" for v in get_file_type_from_directory_recursively(
            f"{path}\\{dir}", file_type))
    res.extend(list(
        filter(
            lambda name: Path(f"{path}\\{name}").suffix == file_type,
            get_files(path)
        )
    ))
    return res


def rename_file(path: str, new_name: str) -> None:
    new_path = "./" + \
        "/".join(Path(path).parts[:-1])+"/"+new_name+Path(path).suffix
    move_file(path, new_path)


def move_file(old_path: str, new_path: str) -> None:
    os.rename(old_path, new_path)


async def open_file(file_path: str, application_path: str):
    import subprocess
    p = subprocess.Popen([application_path, file_path])
    return_code = p.wait()


def move_directory(old_path: str, new_path: str) -> None:
    shutil.move(old_path, new_path)


def copy_file(src: str, dest: str) -> None:
    shutil.copy(src, dest)


def copy_directory(src: str, dest: str) -> None:
    shutil.copy(src, dest)


__all__ = [
    "write_to_file",
    "path_exists",
    "file_exists",
    "directory_exists",
    "delete_file",
    "read_file",
    "is_file",
    "is_directory",
    "get_files",
    "get_files_and_directories",
    "get_directories",
    "delete_directory",
    "clear_directory",
    "create_directory",
    "get_file_type_from_directory",
    "get_file_type_from_directory_recursively",
    "rename_file",
    "move_file",
    "open_file",
    "move_directory",
    "copy_file",
    "copy_directory"
]



from typing import Sequence, Union, Any
from collections.abc import Callable
from .Decorators import validate
from .Colors import ColoredText
from .Classes import DisablePytestDiscovery
from .Functions import isoneof
from .IO import read_file, write_to_file, file_exists
from pathlib import Path
import re


class Test(DisablePytestDiscovery):
    @staticmethod
    def passed(text: str):
        pass_text = ColoredText.green("PASSED")
        print(f"{pass_text}: {text}")

    @staticmethod
    def failed(text: str):
        failed_text = ColoredText.red("FAILED")
        print(f"{failed_text}: {text}")

    def __init__(self, inputs: Union[Sequence, Any], outputs: Union[Sequence, Any] = None, exception=None):
        if outputs is None and exception is None:
            raise ValueError(
                "Cannot create a test where which doesn't return anything and not raises any exception")
        if outputs and exception:
            raise ValueError("cant check both return value and exception")

        self.inputs = inputs if isinstance(inputs, Sequence) else (inputs,)
        self.outputs = None
        if outputs is not None:
            self.outputs = outputs if isoneof(
                outputs, [list, tuple]) else (outputs,)
        self.exceptions = None
        if exception is not None:
            self.exceptions = exception.__qualname__

    def acquire_result(self, func: Callable) -> bool:
        return func(*self.inputs)

    def is_passing(self, func: Callable) -> bool:
        output = func(*self.inputs)
        if not isoneof(output, [list, tuple]):
            output = (output,)
        return output == self.outputs

    def test(self, func: Callable, verbose: bool = False, test_number: int = 0) -> bool:
        has_passed = False
        try:
            msg = None
            res = self.acquire_result(func)
            if not isoneof(res, [list, tuple]):
                res = (res,)
            has_passed = res == self.outputs
            msg = f"{test_number}: {func.__qualname__}{self.inputs} => {res} := {self.outputs}"
        except Exception as e:
            if type(e).__qualname__ == self.exceptions:
                passed_test = True
            msg = f"{test_number}: {func.__qualname__}{self.inputs} => {type(e).__qualname__}"
            if passed_test:
                msg += f" := {self.exceptions}"
            else:
                msg += f"{e.args} := {self.exceptions}"
        finally:
            if has_passed:
                if verbose:
                    Test.passed(msg)
            else:
                Test.failed(msg)
            return has_passed


# @pytest.mark.filterwarnings("ignore:api v1")
class TestFactory(DisablePytestDiscovery):
    @validate(None, Callable, bool)
    def __init__(self, func: Callable, verbose: bool = False):
        self.func = func
        self.tests: list[Test] = []
        self.verbose = verbose

    @validate(None, Test)
    def add_test(self, test: Test):
        self.tests.append(test)
        return self

    def add_tests(self, tests: Sequence[Test]):
        self.tests.extend(tests)
        return self

    def __call__(self) -> bool:
        name = f"{self.func.__module__}.{self.func.__qualname__}"
        print(f"Testing {name}...\n")
        count: int = 0
        pass_count: int = 0
        for i, test in enumerate(self.tests):
            if test.test(self.func, self.verbose, i):
                pass_count += 1
            count += 1
        print()
        print(f"PASSED {pass_count} / {count}".center(20, " ").center(40, "="))
        return pass_count == count

    def execute(self) -> bool:
        return self()


@validate(str, str, bool)
def create_test_file(path: str, output_folder: str = None, overwrite: bool = False):
    filename = Path(path).stem
    output_path = f"./test_{filename.lower()}.py"
    if output_folder is not None:
        output_path = f"{output_folder}/test_{filename.lower()}.py"
    if output_folder is not None and file_exists(output_path) and not overwrite:
        return

    lines = read_file(path)
    allowed = ["def", "class", "@property"]

    def is_ok(line: str):
        for allow in allowed:
            if line.strip().startswith(allow):
                return True
        return False
    lines = filter(is_ok, lines)

    in_file_parts = []
    import_path = ".".join([part for part in Path(path).parts])[:-3]
    res = [
        "from danielutils import TestFactory, Test\n",
        f"from {import_path} import *\n",
        "\n\n"
    ]
    indents = 0
    is_property: bool = False
    for line in lines:
        matches = re.search(r"^( {4})+", line)
        if matches is not None:
            indents = matches.regs[0][1]//4
        else:
            indents = 0
        if indents == 0 and len(in_file_parts) > 0:
            in_file_parts.pop()

        if "@property" in line:
            is_property = True
            continue

        if line.strip().startswith("class"):
            class_name = re.findall(r"class (\b\w+\b)", line)
            if class_name:
                in_file_parts.append(class_name[0])
            continue
        if indents != len(in_file_parts):
            continue

        name = re.findall(r"def (.+?)\(", line.strip())
        if not name or "#" in line or line.startswith("def __"):
            continue
        name = name[0]
        res.append(f"def test_{name}():\n")
        func_name = ".".join(v for v in in_file_parts+[name])

        if is_property:
            params = "*args,**kwargs"
            res.append(
                f"\tdef inner({params}): return {filename}({params}).{name}\n")
            func_name = "inner"
            is_property = False
        res.append(
            f"\tassert TestFactory({func_name}).add_tests([\n\t\n\t])()\n\n\n")
    if len(res) > 3:
        write_to_file(output_path, res)


__all__ = [
    "Test",
    "TestFactory",
    "create_test_file"
]
# def test_func(functor, inputs, outputs) -> None:
#     if not callable(functor):
#         raise TypeError("functor must return true for callable(functor)")

#     if len(inputs) != len(outputs):
#         raise ValueError("Amount of inputs and outputs is different")

#     for input, output in zip(inputs, outputs):
#         res = functor(*input[0], **input[1])

# class Tester:
#     @validate(None, Callable, Sequence, Sequence)
#     def __init__(self, func, inputs, outputs):
#         self.func = func
#         self.inputs = inputs
#         self.outputs = outputs

#     def test():
#         pass



import os


def get_current_working_directory() -> str:
    """return the current working directory

    Returns:
        str: current working directory
    """
    return os.getcwd()


def set_current_working_directory(path: str) -> None:
    """sets the current working directory

    Args:
        path (str): directory to set to
    """
    os.chdir(path)


def get_absolute_path(path: str) -> str:
    """return the absolute path of given path

    Args:
        path (str): a path

    Returns:
        str: absolute version of that path
    """
    return os.path.abspath(path)


def get_relative_path(path: str) -> str:
    """return the relative path of given path

    Args:
        path (str): a path

    Returns:
        str: relative (to current working directory) version of that path
    """
    return os.path.realpath(path)


__all__ = [
    "get_current_working_directory",
    "set_current_working_directory",
    "get_absolute_path"
]



from typing import Iterable, Callable


class frange:
    def __init__(self, start: float, stop: float = None, step: float = 1, round_method: Callable[[float], float] = lambda f: round(f, 3)):
        if stop is None:
            stop = start
            start = 0
        self.start = start
        self.stop = stop
        self.step = step
        self.method = round_method

    def __iter__(self) -> Iterable:
        if self.stop < self.start:
            return
        if self.start > self.stop:
            return
        if abs(self.stop-self.start) < abs(self.step):
            return
        if self.stop > 0 and self.step < 0:
            return
        if self.stop < 0 and self.step > 0:
            return
        cur = self.start
        while cur < self.stop:
            yield self.method(cur)
            cur += self.step


__all__ = [
    "frange"
]



import sys
from ..Functions import areoneof
from .MathSymbols import subscript_dict, superscript_dict


def __mprint_parse_one(s: str) -> str:

    def inner(res, index, dct):
        start = index
        while index < len(s) and s[index] not in {' ', '*', '+', '-', '/', '_', '^'}:
            index += 1
        end = index
        for c in s[start:end]:
            if c in dct:
                res += dct[c]
            else:
                res += c
        index -= 1
        return res, index
    res: str = ""
    i = 0
    while i < len(s):
        c = s[i]
        if c == "^":
            i += 1
            res, i = inner(res, i, superscript_dict)
        elif c == "_":
            i += 1
            res, i = inner(res, i, subscript_dict)
        else:
            res += c
        i += 1
    return res


def mprint(*args: list[str], sep=" ", end="\n", stream=sys.stdout) -> None:
    if not areoneof([*args], [str]):
        raise TypeError("s must be a string")
    stream.write(sep.join([__mprint_parse_one(s) for s in args])+end)


__all__ = [
    "mprint"
]



import os
from math import inf
from .Decorators import overload, timeout, validate
from .Typing import IO
from .Exceptions import TimeoutError
from .Conversions import str_to_bytes
from .Functions import areoneof
from pathlib import Path
import subprocess
import time


def cm(*args, shell: bool = True) -> tuple[int, bytes, bytes]:
    """Execute windows shell command and return output

    Args:
        command or args:\n
        command (str): A string representation of the command to execute.
        args (list[str]): A list of all the command parts
        shell (bool, optional): whether to execute in shell. Defaults to True.

    Raises:
        TypeError: will raise if 'shell' is not boolean

    Returns:
        Tuple[int, bytes, bytes]: return code, stdout, stderr
    """
    if not isinstance(shell, bool):
        raise TypeError(
            "In function 'cm' param 'shell' must be of type bool")
    for i, arg in enumerate(args):
        if Path(args[i]).is_file() or Path(args[i]).is_dir():
            args = (*args[:i], f"\"{arg}\"", *args[i+1:])
    res = subprocess.run(" ".join(args), shell=shell, capture_output=True)
    return res.returncode, res.stdout, res.stderr


@validate([int, float])
def sleep(seconds: int | float):
    """make current thread sleep

    Args:
        seconds (float): number of seconds to sleep
    """
    time.sleep(seconds)


def __acm_write(*args, p: subprocess.Popen, sep=" ", end="\n") -> None:
    b_args = str_to_bytes(sep).join(str_to_bytes(v) for v in args)
    b_end = str_to_bytes(end)
    p.stdin.write(b_args+b_end)
    p.stdin.flush()


@validate(str, [list, lambda l:areoneof(l, [str]), "inputs must be a list of strings"], [int, float], bool, bool, None, None,)
def acm(command: str, inputs: list[str] = None, i_timeout: float = 0.01, shell: bool = False, use_write_helper: bool = True, cwd=None, env=None) -> tuple[int, list[bytes] | None, list[bytes] | None]:
    """Advanced command

    Args:
        command (str): The command to execute\n
        inputs (list[str]): the inputs to give to the program from the command. Defaults to None.\n
        i_timeout (float, optional): An individual timeout for every step of the execution. Defaults to 0.01.\n
        cwd (?, optional): Current working directory. Defaults to None.\n
        env (?, optional): Environment variables. Defaults to None.\n
        shell (bool, optional): whether to execute the command through shell. Defaults to False.\n
        use_write_helper (bool, optional): whether to parse each input as it would have been parse with builtin print() or to use raw text. Defaults to True.

    Raises:
        If @timeout will raise something other than TimeoutError.\n
        If the subprocess input and output handling will raise an exception.

    Returns:
        tuple[int, list[bytes] | None, list[bytes] | None]: return code, stdout, stderr
    """

    if inputs is None:
        inputs = []
    p = None
    try:
        p = subprocess.Popen(command, stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd, env=env, shell=shell)

        @timeout(i_timeout)
        def readlines(s: IO, l: list):
            l.extend(s.readlines())

        def extend_from_stream(stream: IO[bytes], list_to_extend_to: list):
            if stream is not None and stream.readable():
                try:
                    readlines(stream, list_to_extend_to)
                    # new_len = len(l)
                except TimeoutError:
                    # break
                    pass
                except BaseException as e1:
                    raise e1

        stdout: list[bytes] = []
        stderr: list[bytes] = []
        for input in inputs:
            if p.stdin.writable():
                if use_write_helper:
                    __acm_write(input, p=p)
                else:
                    __acm_write(input, p=p, sep="", end="")
            extend_from_stream(p.stdout, stdout)
            extend_from_stream(p.stderr, stderr)
        else:
            extend_from_stream(p.stdout, stdout)
            extend_from_stream(p.stderr, stderr)
        p.stdin.close()
        p.stdout.close()
        if p.stderr is not None:
            p.stderr.close()
        returncode = p.wait()
        return returncode, stdout, stderr
    except BaseException as e2:
        raise type(e2)(f"Maybe use shell=True? original error:\n{e2.args}")
    finally:
        if p is not None:
            if p.stdin is not None:
                p.stdin.close()
            if p.stderr is not None:
                p.stderr.close()
            if p.stdout is not None:
                p.stdout.close()


__all__ = [
    "cm",
    "acm",
    "sleep"
]



from __future__ import annotations
from typing import Any


class Node:
    def __init__(self, data: Any, next=None) -> None:
        self.data = data
        self.next = next

    def __str__(self) -> str:
        return f"Node(data={self.data}, next={self.next})"

    def __repr__(self) -> str:
        return str(self)


class Tree:
    pass


class BinaryTree(Tree):
    pass


class Stack:
    def __init__(self):
        self.data = []

    def pop(self) -> Any:
        if not self.is_empty():
            res = self.data[-1]
            self.data = self.data[:-1]
            return res

    def push(self, v: Any) -> None:
        self.data.append(v)

    def is_empty(self) -> bool:
        return len(self) == 0

    def __str__(self) -> str:
        s = ", ".join(str(v) for v in self.data)
        return f"Stack({s})"

    def __repr__(self) -> str:
        return str(self)

    def __len__(self) -> int:
        return len(self.data)

    def top(self) -> Any:
        return self.data[-1]


class Queue:
    def __init__(self):
        self.data = []

    def dequeue(self) -> Any:
        if not self.is_empty():
            res = self.data[0]
            self.data = self.data[1:]
            return res

    def enqueue(self, v: Any) -> None:
        self.data.append(v)

    def is_empty(self) -> bool:
        return len(self) == 0

    def __str__(self) -> str:
        s = ", ".join(str(v) for v in self.data)
        return f"Queue({s})"

    def __repr__(self) -> str:
        return str(self)

    def __len__(self) -> int:
        return len(self.data)


class LinkedList:
    pass



from .Decorators import validate


class ColoredText:
    __rgb_validation = [int, lambda v: 0 <=
                         v <= 255, "rgb value must be in [0,255]"]

    @staticmethod
    @validate(__rgb_validation, __rgb_validation, __rgb_validation, str)
    def from_rgb(r: int, g: int, b: int, text: str):
        return f"\033[38;2;{r};{g};{b}m{text}\033[38;2;255;255;255m"

    @staticmethod
    def green(text: str):
        return ColoredText.from_rgb(0, 255, 0, text)

    @staticmethod
    def blue(text: str):
        return ColoredText.from_rgb(0, 0, 255, text)

    @staticmethod
    def red(text: str):
        return ColoredText.from_rgb(255, 0, 0, text)

    @staticmethod
    def yellow(text: str):
        return ColoredText.from_rgb(255, 255, 0, text)

    @staticmethod
    def white(text: str):
        return ColoredText.from_rgb(255, 255, 255, text)

    @staticmethod
    def black(text: str):
        return ColoredText.from_rgb(0, 0, 0, text)


def __special_print(*args, sep=" ", end="\n", start_with=None):
    if start_with:
        if "\n" not in sep:
            print(f"{start_with}: ", end="")
            print(sep.join([arg for arg in args]), sep="", end=end)
        else:
            print(
                sep.join([f"{start_with}: {arg}" for arg in args]), sep="", end=end)
    else:
        print(*args, sep=sep, end=end)


def warning(*args, sep=" ", end="\n"):
    """print a warning message

    Args:
        sep (str, optional): print separator. Defaults to " ".
        end (str, optional): print endline. Defaults to "\\n".
    """

    __special_print(*args, sep=sep, end=end,
                    start_with=ColoredText.yellow("WARNING"))


def error(*args, sep=" ", end="\n"):
    """print an error message

    Args:
        sep (str, optional): print separator. Defaults to " ".
        end (str, optional): print endline. Defaults to "\\n".
    """
    __special_print(*args, sep=sep, end=end,
                    start_with=ColoredText.red("ERROR"))


__all__ = [
    "ColoredText",
    "warning",
    "error"
]



from __future__ import annotations
from ..Typing import Any, Sequence, Union, TypeVar
from ..Decorators import overload, validate


class GraphNode:
    def __init__(self, value: Any):
        self.value = value

    def __str__(self) -> str:
        return f"GN(V={self.value})"

    def __repr__(self) -> str:
        return str(self)


class Connection:
    @validate(None, GraphNode, GraphNode, Any)
    def __init__(self, node1: GraphNode, node2: GraphNode, weight: Any):
        self.node1 = node1
        self.node2 = node2
        self.weight = weight


class Graph:
    @classmethod
    @validate(None, list, list, list)
    def from_lists(cls, values: list[Any], connections: list[list[int]]) -> Graph:
        nodes = [GraphNode(v) for v in values]
        parsed_connections = []
        for i1, i_connections in enumerate(connections):
            for i2 in i_connections:
                parsed_connections.append(Connection(nodes[i1], nodes[i2]))
        return cls(
            nodes,
            parsed_connections,
        )

    def __init__(self, nodes: list[GraphNode], connections: list[Connection]):
        if not (len(nodes) == len(connections)):
            raise ValueError("Length of input params is not th same")
        self.nodes = nodes
        self.connections = connections

    def __str__(self) -> str:
        res = ", ".join([
            f"({value},{connections},{weights})" for value, connections, weights in self
        ])
        return res

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self):
        for i in range(len(self)):
            yield (self.nodes[i], self.connections[i], self.weights[i]if self.weights is not None else None)

    @overload(None, GraphNode)
    def __contains__(self, node: GraphNode) -> bool:
        return node in self.nodes

    @overload(None, Connection)
    def __contains__(self, connection: Connection) -> bool:
        return connection in self.connections

    def clone() -> Graph:
        pass


class TypedGraph(Graph):
    def __init__(self, nodes: list[GraphNode], connections: list[list[int]], weights: list[list[Any]] = None) -> None:
        for i in range(1, len(nodes)):
            if type(nodes[0].value) != type(nodes[i].value):
                raise TypeError(
                    "One or more of the values is of different type than the others")
        super().__init__(nodes, connections, weights)


__all__ = [
    "Graph",
    "TypedGraph"
]



from .Decorators import atomic

# although it's related to threading, atomic decorator implemented at .Decorators


@atomic
def atomic_print(*args, **kwargs):
    """exactly the same params and behavior as builtin print(), but behaves atomic-ly
    """
    print(*args, **kwargs)


__all__ = [
    "atomic_print"
]



# https://unicode-table.com/en/sets/mathematical-signs/
special_double_n = "ℕ"
special_double_q = "ℚ"
special_double_r = "ℝ"
special_double_z = "ℤ"
special_double_c = "ℂ"
special_rpab = "〉"  # right pointing angle bracket
special_lpab = "〈"  # left pointing angle bracket
special_forall = "∀"
special_exists = "∃"
special_vl = "|"
special_sigma = "∑"
special_pi = "∏"
special_circled_plus = "⊕"
special_ne = "≠"
special_equiv = "≡"
special_lambda = "λ"
special_square_root = "√"
special_cube_root = "∛"
special_forth_root = "∜"


superscript_small_case_a = 'ᵃ'
superscript_small_case_b = 'ᵇ'
superscript_small_case_c = 'ᶜ'
superscript_small_case_d = 'ᵈ'
superscript_small_case_e = 'ᵉ'
superscript_small_case_f = 'ᶠ'
superscript_small_case_g = 'ᵍ'
superscript_small_case_h = 'ʰ'
superscript_small_case_i = 'ⁱ'
superscript_small_case_j = 'ʲ'
superscript_small_case_k = 'ᵏ'
superscript_small_case_l = 'ˡ'
superscript_small_case_m = 'ᵐ'
superscript_small_case_n = 'ⁿ'
superscript_small_case_o = 'ᵒ'
superscript_small_case_p = 'ᵖ'
superscript_small_case_q = '𐞥'
superscript_small_case_r = 'ʳ'
superscript_small_case_s = 'ˢ'
superscript_small_case_t = 'ᵗ'
superscript_small_case_u = 'ᵘ'
superscript_small_case_v = 'ᵛ'
superscript_small_case_w = 'ʷ'
superscript_small_case_x = 'ˣ'
superscript_small_case_y = 'ʸ'
superscript_small_case_z = 'ᶻ'

superscript_small_letters = [
    superscript_small_case_a, superscript_small_case_b, superscript_small_case_c,
    superscript_small_case_d, superscript_small_case_e, superscript_small_case_f,
    superscript_small_case_g, superscript_small_case_h, superscript_small_case_i,
    superscript_small_case_j, superscript_small_case_k, superscript_small_case_l,
    superscript_small_case_m, superscript_small_case_n, superscript_small_case_o,
    superscript_small_case_p, superscript_small_case_q, superscript_small_case_r,
    superscript_small_case_s, superscript_small_case_t, superscript_small_case_u,
    superscript_small_case_v, superscript_small_case_w, superscript_small_case_x,
    superscript_small_case_y, superscript_small_case_z
]

# superscript_big_case_a = 'ⁱ'
# superscript_big_case_b = 'ⁱ'
# superscript_big_case_c = 'ⁱ'
# superscript_big_case_d = 'ⁱ'
# superscript_big_case_e = 'ⁱ'
# superscript_big_case_f = 'ⁱ'
# superscript_big_case_g = 'ⁱ'
# superscript_big_case_h = 'ⁱ'
# superscript_big_case_i = 'ⁱ'
# superscript_big_case_j = 'ⁱ'
# superscript_big_case_k = 'ⁱ'
# superscript_big_case_l = 'ⁱ'
# superscript_big_case_m = 'ⁱ'
# superscript_big_case_n = 'ⁱ'
# superscript_big_case_o = 'ⁱ'
# superscript_big_case_p = 'ⁱ'
# superscript_big_case_q = 'ⁱ'
# superscript_big_case_r = 'ⁱ'
# superscript_big_case_s = 'ⁱ'
# superscript_big_case_t = 'ⁱ'
# superscript_big_case_u = 'ⁱ'
# superscript_big_case_v = 'ⁱ'
# superscript_big_case_w = 'ⁱ'
# superscript_big_case_x = 'ⁱ'
# superscript_big_case_y = 'ⁱ'
# superscript_big_case_z = 'ⁱ'


superscript_dict = dict()
superscript_dict.update(
    {chr(i+ord('a')): superscript_small_letters[i] for i in range(26)}
)
superscript_digits = ["⁰", "¹", "²", "³",
                      "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"]
superscript_dict.update(
    {i: superscript_digits[i] for i in range(len(superscript_digits))}
)
superscript_dict.update({
    "+": "⁺",
    "-": "⁻",
    "=": "⁼",
    "(": "⁽",
    ")": "⁾",
})

subscript_small_case_a = 'ₐ'
subscript_small_case_b = 'ᵇ'
subscript_small_case_c = ''
subscript_small_case_d = 'ᵈ'
subscript_small_case_e = 'ₑ'
subscript_small_case_f = ''
subscript_small_case_g = 'ᶢ'
subscript_small_case_h = 'ₕ'
subscript_small_case_i = 'ᵢ'
subscript_small_case_j = 'ⱼ'
subscript_small_case_k = 'ₖ'
subscript_small_case_l = 'ₗ'
subscript_small_case_m = 'ₘ'
subscript_small_case_n = 'ₙ'
subscript_small_case_o = 'ₒ'
subscript_small_case_p = 'ₚ'
subscript_small_case_q = ''
subscript_small_case_r = 'ᵣ'
subscript_small_case_s = 'ₛ'
subscript_small_case_t = 'ₜ'
subscript_small_case_u = 'ᵤ'
subscript_small_case_v = 'ᵥ'
subscript_small_case_w = ''
subscript_small_case_x = 'ₓ'
subscript_small_case_y = ''
subscript_small_case_z = ''
subscript_small_letters = [subscript_small_case_a, subscript_small_case_b, subscript_small_case_c, subscript_small_case_d, subscript_small_case_e, subscript_small_case_f, subscript_small_case_g, subscript_small_case_h, subscript_small_case_i, subscript_small_case_j, subscript_small_case_k, subscript_small_case_l,
                           subscript_small_case_m, subscript_small_case_n, subscript_small_case_o, subscript_small_case_p, subscript_small_case_q, subscript_small_case_r, subscript_small_case_s, subscript_small_case_t, subscript_small_case_u, subscript_small_case_v, subscript_small_case_w, subscript_small_case_x, subscript_small_case_y, subscript_small_case_z]
# subscript_big_case_a = 'ⁱ'
# subscript_big_case_b = 'ⁱ'
# subscript_big_case_c = 'ⁱ'
# subscript_big_case_d = 'ⁱ'
# subscript_big_case_e = 'ⁱ'
# subscript_big_case_f = 'ⁱ'
# subscript_big_case_g = 'ⁱ'
# subscript_big_case_h = 'ⁱ'
# subscript_big_case_i = 'ⁱ'
# subscript_big_case_j = 'ⁱ'
# subscript_big_case_k = 'ⁱ'
# subscript_big_case_l = 'ⁱ'
# subscript_big_case_m = 'ⁱ'
# subscript_big_case_n = 'ⁱ'
# subscript_big_case_o = 'ⁱ'
# subscript_big_case_p = 'ⁱ'
# subscript_big_case_q = 'ⁱ'
# subscript_big_case_r = 'ⁱ'
# subscript_big_case_s = 'ⁱ'
# subscript_big_case_t = 'ⁱ'
# subscript_big_case_u = 'ⁱ'
# subscript_big_case_v = 'ⁱ'
# subscript_big_case_w = 'ⁱ'
# subscript_big_case_x = 'ⁱ'
# subscript_big_case_y = 'ⁱ'
# subscript_big_case_z = 'ⁱ'
subscript_dict = dict()
subscript_dict.update(
    {chr(i+ord('a')): subscript_small_letters[i]
     for i in range(len(subscript_small_letters))}
)

subscript_digits = ["\u2080", "\u2081", "\u2082", "\u2083",
                    "\u2084", "\u2085", "\u2086", "\u2087", "\u2088", "\u2089"]
subscript_dict.update(
    {f'{i}': subscript_digits[i] for i in range(len(subscript_digits))}
)
subscript_dict.update({
    "+": "\u208A",
    "-": "\u208B",
    "=": "\u208C",
    "(": "\u208D",
    ")": "\u208E",
})
__all__ = [
    "subscript_dict",
    "subscript_dict",
]



from .Typing import Any, type, Sequence, Union, Iterable, Callable
from typing import get_args, get_origin, get_type_hints


def __isoftype_inquire(obj: Any) -> tuple[Any, Any, Any]:
    origin = None
    args = None
    type_hints = None
    try:
        origin = get_origin(obj)
    except:
        pass
    try:
        args = get_args(obj)
    except:
        pass
    try:
        type_hints = get_type_hints(obj)
    except:
        pass
    return origin, args, type_hints


def isoftype(obj: Any, T: Any, /, strict: bool = True) -> bool:
    """Checks if an object is of the given type or any of its subtypes.

    Args:
        obj (Any): The object to be checked.
        T (Any): The type or types to be checked against.

    Returns:
        bool: True if the object is of the given type or any of its subtypes, False otherwise.
    """
    if not isinstance(strict, bool):
        raise TypeError("'strict' must be of type bool")
    obj_origin, obj_args, obj_hints = __isoftype_inquire(obj)
    t_origin, t_args, t_hints = __isoftype_inquire(T)
    if t_origin is not None:
        if t_origin in {list}:
            for sub_v in obj:
                if not isoftype(sub_v, t_args[0]):
                    return False
            return True
        elif t_origin is dict:
            key_t, value_t,  = t_args[0], t_args[1]
            for k, v in obj.items():
                if not isoftype(v, value_t):
                    return False
                if not isoftype(k, key_t):
                    return False
            return True
        elif t_origin in {Union}:
            for sub_t in t_args:
                if isoftype(obj, sub_t):
                    return True
            return False
        elif t_origin in {Callable}:
            if obj.__name__ == "<lambda>":
                print("using lambda function with isoftype is ambiguous")
                return not strict
            tmp = list(obj_hints.values())
            obj_return_type = tmp[-1]
            obj_param_types = tuple(tmp[:-1])
            del tmp
            if len(t_args) == 0:
                return True
            t_return_type = t_args[1]
            t_param_types = tuple(t_args[0])
            return obj_return_type is t_return_type and obj_param_types == t_param_types
    else:
        if T is Any:
            return True
        elif type(T) in {list}:
            for sub_t in T:
                if isoftype(obj, sub_t):
                    return True
            return False
        elif obj_origin is not None:
            if obj_origin is Union:
                return T is type(Union)
    return isinstance(obj, T)


# def isoftype(v: Any, t: Any):
#     if isinstance(v, (list, tuple)):
#         if hasattr(t, '__args__'):
#             for sub_v in v:
#                 if not isoftype(sub_v, t.__args__[0]):
#                     return False
#             return True
#         else:
#             if t in (list, tuple, Iterable):
#                 return True
#             return False
#     elif isinstance(v, dict):
#         if hasattr(t, '__args__'):
#             key_t, value_t = t.__args__[0], t.__args__[1]
#             for key, value in v.items():
#                 if not isoftype(key, key_t) or not isoftype(value, value_t):
#                     return False
#             return True
#         else:
#             if t is dict:
#                 return True
#             return False
#     else:  # for instances but also for Union
#         if t.__module__ == "typing":
#             if t == Any:
#                 return True
#             elif hasattr(v, '__origin__'):
#                 if v.__origin__ == Union:
#                     return type(Union) == t
#         elif "Callable" in str(t):
#             if v is not Callable and "Callable" in str(v):
#                 # the case is Callable[[somethings],something]
#                 if t is type(Callable):
#                     return True
#                 # v_args = v.__args__
#                 # t_args = t.__args__
#                 pass
#         elif hasattr(t, '__origin__'):
#             if isinstance(t.__origin__, list):
#                 return False
#         elif type(t) == list or type(t) == tuple:
#             return isoneof(v, t)
#         return isinstance(v, t)


def isoneof(v: Any, types: Union[list[type], tuple[type]]) -> bool:
    """performs isoftype() or ... or isoftype()

    Args:
        v (Any): the value to check it's type
        types (Union[list[Type], tuple[Type]): A Sequence of approved types

    Raises:
        TypeError: if the second argument is not from Union[list[Type], tuple[Type]

    Returns:
        bool: return True iff isoftype(v, types[0]) or ... isoftype(v, types[...])
    """
    if not isinstance(types, (list, tuple)):
        raise TypeError("'types' must be of type 'list' or 'tuple'")
    for T in types:
        if isoftype(v, T):
            return True
    return False


def isoneof_strict(v: Any, types: Sequence[type]) -> bool:
    """performs 'type(v) in types' efficiently

    Args:
        v (Any): value to check
        types (Sequence[Type]): sequence of approved types

    Raises:
        TypeError: if types is not a sequence

    Returns:
        bool: true if type of value appears in types
    """
    if not isinstance(types, Sequence):
        raise TypeError("lst must be of type Sequence")
    for T in types:
        if type(v) == T:
            return True
    return False


def areoneof(values: Sequence[Any], types: Sequence[type]) -> bool:
    """performs 'isoneof(values[0],types) and ... and isoneof(values[...],types)'

    Args:
        values (Sequence[Any]): Sequence of values
        types (Sequence[Type]): Sequence of types

    Raises:
        TypeError: if types is not a Sequence
        TypeError: if values is not a Sequence

    Returns:
        bool: the result of the check
    """
    if not isinstance(types, Sequence):
        raise TypeError("'types' must be of type Sequence")
    if not isinstance(values, Sequence):
        raise TypeError("'values' must be of type Sequence")
    for v in values:
        if not isoneof(v, types):
            return False
    return True


def check_foreach(values: Sequence[Any], condition: Callable[[Any], bool]) -> bool:
    """

    Args:
        values (Sequence[Any]): Values to perform check on
        condition (Callable[[Any], bool]): Condition to check on all values

    Returns:
        bool: returns True iff condition return True for all values individually
    """
    if not isinstance(values, Sequence):
        pass
    if not isinstance(condition, Callable):
        pass
    for v in values:
        if not condition(v):
            return False
    return True


def get_source_code(obj: object) -> str:
    if not isinstance(obj, object):
        raise ValueError("obj must be an object")
    import inspect
    return inspect.getsource(obj)


__all__ = [
    "isoneof",
    "isoneof_strict",
    "areoneof",
    "check_foreach",
    "isoftype",
    "get_source_code"
]
# def almost_equal(*args: Sequence[Any], func: Callable[[Any, Any, Any], bool] = math.isclose, diff: Any = 0.00000000001) -> bool:
#     """checks whether all values are within absolute range of each other in O(n**2)

#     Args:
#         func (Callable[[Any, Any, Any], bool], optional): function to check. Defaults to math.isclose.
#         diff (Any, optional): default absolute tolerance. Defaults to 0.00000000001.

#     Returns:
#         bool: return True if all values are within specified tolerance from all other values
#     """
#     for i in range(len(args)):
#         for j in range(i+1, len(args)):
#             if func is math.isclose:
#                 if not func(args[i], args[j], abs_tol=diff):
#                     return False
#             else:
#                 if not func(args[i], args[j], diff):
#                     return False
#     return True



from screeninfo import get_monitors
import pyautogui
from PIL import ImageGrab, Image
import PIL
from functools import partial
ImageGrab.grab = partial(ImageGrab.grab, all_screens=True)


def screenshot(xy: tuple[int, int] = (0, 0), wh: tuple[int, int] = None) -> Image:
    if wh is None:
        wh = (0, 0)
        for monitor_index in range(get_monitor_count()):
            size = get_monitor_size(monitor_index)
            wh = (wh[0]+size[0], wh[1]+size[1])
    return pyautogui.screenshot(None, (xy[0], xy[1], wh[0], wh[1]))


def screenshot_monitor(index: int) -> Image:
    return screenshot(get_monitor_location(index), get_monitor_size(index))


def get_monitor_size(index: int) -> tuple[int, int]:
    monitor = get_monitors()[index]
    return monitor.width, monitor.height


def get_monitor_location(index) -> tuple[int, int]:
    monitor = get_monitors()[index]
    return monitor.x, monitor.y


def get_monitor_count() -> int:
    """return the number of displays connected to this computer

    Returns:
        int: amount of displays
    """
    return len(get_monitors())


__all__ = [
    "screenshot",
    "get_monitor_size",
    "get_monitor_count",
    "screenshot_monitor"
]



import json
from .Decorators import validate


@validate(dict)
def dict_to_json(d: dict) -> str:
    return json.dumps(d, indent=4)


@validate(str)
def json_to_dict(j: str) -> dict:
    return json.loads(j)


__all__ = [
    "dict_to_json",
    "json_to_dict"
]



import urllib.request
import urllib.parse
from urllib.parse import urlparse
import urllib
from .Decorators import validate


def prettify_html(html: str) -> str:
    return html


@validate(str)
def get_html(url: str) -> str:
    user_agent = 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.0.7) Gecko/2009021910 Firefox/3.0.7'
    headers = {'User-Agent': user_agent, }
    req = urllib.request.Request(url, headers=headers)
    html = urllib.request.urlopen(req).read().decode('UTF-8')
    # return bs4(html, 'html.parser').prettify()
    return html


@validate(str)
def get_url_details(url: str) -> tuple[str, str, str, str, str]:
    scheme, netloc, path, params, query, fragment = urlparse(url)
    return scheme, netloc, path, params, query, fragment


@validate(str)
def url_encode(s: str) -> str:
    return urllib.parse.quote(s)


@validate(str)
def url_decode(s: str) -> str:
    return urllib.parse.unquote_plus(s)


def get_elements(html: str, tag: str) -> list[str]:
    from bs4 import BeautifulSoup as bs4
    return [str(v) for v in bs4(html, 'html.parser').find_all(tag)]


__all__ = [
    "prettify_html",
    "get_html",
    "get_url_details",
    "url_encode",
    "url_decode",
    "get_elements"
]



from .Functions import *
from .Decorators import *
from .IO import *
from .System import *
from .Text import *
from .Conversions import *
from .Classes import *
from .Testing import *
from .Time import *
from .Internet import *
from .Threading import *
from .Colors import *
from .Data import *
from .Typing import *
from .DataStructures import *
from .Math import *
from .Serialization import *
from .Path import *
from .Windows import *


from __future__ import annotations
from ..Typing import Any, Union, Iterable
from ..Decorators import validate, overload
from ..Functions import isoftype


class tlist(list):
    """tlist is same as builtin python list but with added type restriction

    Args:
        type (type): the allowed type, can be nested type
        iterable (Iterable, optional): the value to create the tlist from. Defaults to None.
    """

    def get_type_error_msg(self, v: Any):
        return f"A value is of the wrong type:\n'{v}' is of type '{type(v)}' but should be of type '{self.type}'"

    @validate(None, [type, type(Union[Any, Any])], Iterable)
    def __init__(self, type: type, iterable: Iterable = None):
        """_summary_

        Args:
            type (type): the allowed type, can be nested type
        ietrable (Iterable, optional): the value to create the tlist from. Defaults to None.

        Raises:
            TypeError: _description_
        """
        self.type = type
        if iterable is not None:
            for v in iterable:
                if not isoftype(v, type):
                    raise TypeError(self.get_type_error_msg(v))
                super().append(v)

    def __setitem__(self, index: int, value: Any) -> None:
        if not isoftype(value, self.type):
            raise TypeError(self.get_type_error_msg(value))
        super()[index] = value

    def append(self, value: Any) -> None:
        if not isoftype(value, self.type):
            raise TypeError(self.get_type_error_msg(value))
        super().append(value)

    @validate(None, Iterable)
    def extend(self, iterable: Iterable) -> None:
        for v in iterable:
            self.append(v)

    def __add__(self, other) -> tlist:
        pass


class tdict(dict):
    @overload(None, type, type)
    def __init__(self, key_t: type, val_t: type):
        self.key_t = key_t
        self.val_t = val_t
        super().__init__()

    @overload(None, type, type, Iterable)
    def __init__(self, keyt: type, val_t: type, iterable: Iterable[tuple]):
        self.key_t = keyt
        self.val_t = val_t
        super().__init__(iterable)

    @overload(None, type, type, dict)
    def __init__(self, keyt: type, val_t: type, ** kwargs):
        """dict(type,type) -> new empty dictionary dict(mapping) -> new dictionary initialized from a mapping object's
    (key, value) pairs
dict(type,type,iterable) -> new dictionary initialized as if via:
    d = {} for k, v in iterable:
        d[k] = v
dict(type,type,**kwargs) -> new dictionary initialized with the name=value pairs
    in the keyword argument list. For example: dict(one=1, two=2)
        """
        self.key_t = keyt
        self.val_t = val_t
        super().__init__(**kwargs)

    def __setitem__(self, key, value) -> None:
        if not isoftype(key, self.key_t):
            raise TypeError(
                f"In class 'tdict' error creating new key-value pair as key = '{key}' is not of type '{self.key_t}'")
        if not isoftype(value, self.val_t):
            raise TypeError(
                f"In class 'tdict' error creating new key-value pair as value = '{value}' is not of type '{ self.val_t}'")
        super().__setitem__(key, value)

    def __str__(self):
        return f"dict[{self.key_t.__name__}, {self.val_t.__name__}]: {super().__str__()}"


# TODO tset
class tset(set):
    def __init__(self):
        pass


__all__ = [
    "tlist",
    "tdict"
]



from typing import Any, Iterable, Optional, NoReturn, TypeAlias, Union, Concatenate, Literal, ClassVar, Final, Annotated, TypeGuard, Generic, TypeVar, ParamSpec, NewType, IO
from collections.abc import Callable, Sequence
from builtins import tuple, list, dict, type

__all__ = [
    "Any",
    "Iterable",
    "Optional",
    "NoReturn",
    "TypeAlias",
    "Union",
    "Concatenate",
    "Literal",
    "ClassVar",
    "Final",
    "Annotated",
    "TypeGuard",
    "Generic",
    "TypeVar",
    "ParamSpec",
    "NewType",
    "IO",
    "Callable",
    "Sequence",
    "tuple",
    "list",
    "dict",
    "type",
]



from ...Decorators import overload
from ..MainConversions import *


@overload(int)
def to_hex(v: int) -> str:
    # docstring at last implementation
    return int_to_hex(v)


@overload(str)
def to_hex(v: str) -> str:
    """to_hex has several options:\n
    1. type(v) == int\n
    2. type(v) == str and len(v) == 1

    Returns:
        str: str of the hex value
    """
    return char_to_hex(v)


__all__ = [
    "to_hex"
]



class OverloadException(Exception):
    """Base exception for overload decorator
    """
    pass


class OverloadNotFound(OverloadException):
    """
    Exception to raise if a function is called with certain argument types but this function hasn't been overloaded with those types
    """
    pass


class OverloadDuplication(OverloadException):
    """
    Exception to raise if a function is overloaded twice with same argument types
    """
    pass


class ValidationError(Exception):
    pass


class ValidationTypeError(ValidationError, TypeError):
    pass


class ValidationReturnTypeError(ValidationError, TypeError):
    pass


class ValidationValueError(ValidationError, ValueError):
    pass


class ValidationDuplicationError(ValidationError):
    pass


class TimeoutError(AssertionError):
    pass



from .Constants import *
from .MathPrint import *
from .Functions import *
# from .Sigma import *
# from .Polynomial import *
# from .Fraction import *
# from .Limit import *
# from .Expression import *
# from .long import *



from ..Decorators import validate, overload
from ..Typing import Any, Iterable, Union
import random
INDEX_INDEX = 0
OCCUPANCE_INDEX = 1


class MarkovNode:
    @overload(None, Any)
    def __init__(self, value: Any) -> None:
        self.value = value
        self.occurrences = []
        self.indices = []

    @overload(None, Any, list[tuple[int, int]])
    def __init__(self, value: Any, children: list[tuple[float, int]] = None) -> None:
        self.value = value
        self.occurrences = [tup[OCCUPANCE_INDEX] for tup in children]
        self.indices = [tup[INDEX_INDEX] for tup in children]

    @overload(None, Any, list[int], list[int])
    def __init__(self, value: Any,  indices: list[int], occurrences: list[int]) -> None:
        if len(occurrences) != len(indices):
            raise ValueError("")
        self.value = value
        self.occurrences = occurrences
        self.indices = indices

    def __str__(self) -> str:
        return f"({str(self.value)})"

    def add_child(self, occurrence: float, index: int) -> None:
        self.occurrences.append(occurrence)
        self.indices.append(index)

    def next(self) -> int:
        count = sum(self.occurrences)
        selector = random.randint(0, count)
        for i, child_occur in enumerate(self.occurrences):
            selector -= child_occur
            if selector <= 0:
                break
        return self.indices[i]

    def __repr__(self) -> str:
        return str(self)

    def has_children(self) -> bool:
        return len(self.occurrences) > 0


class MarkovChain:
    @validate(None, list[MarkovNode])
    def __init__(self, nodes: list[MarkovNode] = None) -> None:
        self.nodes = nodes if nodes is not None else []

    @validate(None, int)
    def __getitem__(self, index: int) -> MarkovNode:
        return self.nodes[index]

    @validate(None, MarkovNode)
    def append(self, node: MarkovNode) -> None:
        self.nodes.append(node)

    def __iter__(self) -> Iterable:
        return iter(self.nodes)

    def __str__(self) -> str:
        return " -> ".join([str(node) for node in self])

    def __len__(self) -> int:
        return len(self.nodes)

    def generate_sequence(self, max_length: int) -> list[MarkovNode]:
        res = []
        index = 0
        for _ in range(max_length):
            n = self[index]
            res.append(n)
            if not n.has_children():
                break
            index = n.next()
        return res


__all__ = [
    "MarkovNode",
    "MarkovChain"
]



# -*- coding: utf-8 -*-
from .Decorators import validate
from .Functions import check_foreach
from .Typing import Union, TypeGuard
HEBREW_LETTERS = ['\u05D0', '\u2135', '\uFB21', '\uFB2E', '\uFB2F', '\uFB30', '\uFB4F', '\u05D1', '\u2136', '\uFB31', '\uFB4C', '\u05D2', '\u2137', '\uFB32', '\u05D3', '\u2138', '\uFB22', '\uFB33', '\u05D4', '\uFB23', '\uFB34', '\u05D5', '\uFB4B', '\uFB35', '\u05F0', '\u05F1', '\u05D6', '\uFB36', '\u05D7', '\u05D8', '\uFB38', '\u05D9', '\uFB1D', '\uFB39', '\u05EF', '\u05F2', '\uFB1F', '\u05DB', '\uFB24',
                  '\u05DA', '\uFB3B', '\uFB3A', '\uFB4D', '\u05DC', '\uFB25', '\uFB3C', '\u05DE', '\uFB26', '\u05DD', '\uFB3E', '\u05E0', '\u05DF', '\uFB40', '\u05E1', '\uFB41', '\u05E2', '\uFB20', '\u05E4', '\u05E3', '\uFB44', '\uFB43', '\uFB4E', '\u05E6', '\u05E5', '\uFB46', '\u05E7', '\uFB47', '\u05E8', '\uFB27', '\uFB48', '\u05E9', '\uFB2B', '\uFB2A', '\uFB49', '\uFB2D', '\uFB2C', '\u05EA', '\uFB28', '\uFB4A']
HEBREW_LETTERS_DEC = [ord(v) for v in HEBREW_LETTERS]
HEBREW_LETTERS_HEX = [hex(v) for v in HEBREW_LETTERS_DEC]
ENGLISH_LETTERS = [chr(v) for v in range(65, 91)]+[chr(v)
                                                   for v in range(97, 123)]
ENGLISH_LETTERS_DEC = [ord(v) for v in ENGLISH_LETTERS]
ENGLISH_LETTERS_HEX = [hex(v) for v in ENGLISH_LETTERS_DEC]


@validate(str, return_type=bool)
def is_english(s: str) -> TypeGuard[str]:
    return check_foreach(s, lambda c: c in ENGLISH_LETTERS)
    # try:
    #     s.encode(encoding='utf-8').decode('ascii')
    # except UnicodeDecodeError:
    #     return False
    # else:
    #     return True


@validate(str, return_type=bool)
def is_str_number(text: str) -> bool:
    return text.isnumeric()


@validate([int, float])
def is_int(num: Union[int, float]) -> TypeGuard[int]:
    if isinstance(num, int):
        return True

    return int(num) == num


@validate(str, return_type=bool)
def is_float(text: str) -> TypeGuard[float]:
    try:
        float(text)
        return True
    except ValueError:
        return False


@validate(str, return_type=bool)
def is_number(text: str) -> bool:
    return is_float(text)


@validate(str, return_type=bool)
def is_hebrew(text: str) -> TypeGuard[str]:
    return check_foreach(text, lambda c: c in HEBREW_LETTERS)


@validate(str, return_type=bool)
def is_binary(text: str) -> bool:
    return check_foreach(text, lambda c: c in [0, 1])


@validate(str, return_type=bool)
def is_decimal(text: str) -> bool:
    return check_foreach(text, lambda c: c in range(10))


@validate(str, return_type=bool)
def is_hex(h: str) -> bool:
    try:
        int(h, 16)
        return True
    except ValueError:
        return False


__all__ = [
    "HEBREW_LETTERS",
    "HEBREW_LETTERS_DEC",
    "HEBREW_LETTERS_HEX",
    "ENGLISH_LETTERS",
    "ENGLISH_LETTERS_DEC",
    "ENGLISH_LETTERS_HEX",
    "is_english",
    "is_str_number",
    "is_int",
    "is_float",
    "is_number",
    "is_hebrew",
    "is_binary",
    "is_decimal",
    "is_hex"
]



import time


def measure(func, *args, **kwargs):
    start = time.time()
    func(*args, **kwargs)
    end = time.time()
    return end-start


def sleep(secs: float):
    time.sleep(secs)


__all__ = [
    "measure",
    "sleep"
]



from ...Decorators import overload
from ..MainConversions import *
from typing import Union


def to_int(value: str) -> Union[int, list[int]]:
    """converts a single character or a full string to an int or list of int respectively
    """
    if len(value) == 1:
        return char_to_int(value)
    return [char_to_int(ch) for ch in value]


__all__ = [
    "to_int"
]



from setuptools import setup, find_packages
import codecs


def read_file(path: str) -> "list[str]":
    with codecs.open(path, 'r', 'utf-8') as f:
        return [l.strip() for l in f.readlines()]


README_PATH = 'README.md'
DESCRIPTION = 'A python utils library for things I find useful'
VERSION = "0.7.8"
LONG_DESCRIPTION = '\n'.join(read_file(README_PATH))
setup(
    name="danielutils",
    version=VERSION,
    author="danielnachumdev",
    author_email="<danielnachumdev@gmail.com>",
    description=DESCRIPTION,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/danielnachumdev/danielutils',
    license="MIT License",
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "archive/"]),
    install_requires=["bs4", "pyautogui", "screeninfo", "Pillow"],
    platforms=["All"],
    keywords=['functions', 'decorators', 'methods'],
    classifiers=[
        # "Development Status :: In development",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)
# python .\setup.py sdist
# twine upload dist/...



from ..Decorators import validate, overload


@validate([str, lambda s:len(s) == 1, "len(s) must be 1"])
def char_to_int(c: str) -> int:
    """convert char to its representing int value

    Args:
        c (str): char to convert

    Returns:
        int: int value
    """
    return ord(c)


@validate(int)
def int_to_char(num: int) -> str:
    """convert int to its corresponding char

    Args:
        num (int): number to convert

    Returns:
        str: result character
    """
    return chr(num)


@validate(str)
def hex_to_char(h: str) -> str:
    """convert hex number to char

    Args:
        h (str): number to convert

    Returns:
        str: result char
    """
    return int_to_char(hex_to_dec(h))


@validate(str)
def hex_to_dec(h: str) -> int:
    """convert hex to dec

    Args:
        h (str): converts base 16 to base 10

    Returns:
        int: decimal value for hex number
    """
    return int(h, 16)


@validate([str, lambda s:len(s) == 1, "len(s) must be 1"])
def char_to_hex(c: str) -> str:
    """convert char to hex

    Args:
        c (str): char to convert

    Returns:
        str: hex representation
    """
    return int_to_hex(char_to_int(c))


@validate(int)
def dec_to_hex(num: int) -> str:
    """convert decimal number to hex

    Args:
        num (int): number to convert

    Returns:
        str: _description_
    """
    return int_to_hex(num)


@validate(int)
def int_to_hex(num: int) -> str:
    return hex(num)


@validate(bytes)
def bytes_to_str(b: bytes) -> str:
    return b.decode("utf-8")


@validate(str)
def str_to_bytes(s: str) -> bytes:
    return bytes(s, encoding='utf-8')


__all__ = [
    "char_to_int",
    "int_to_char",
    "hex_to_char",
    "hex_to_dec",
    "char_to_hex",
    "dec_to_hex",
    "int_to_hex",
    "bytes_to_str",
    "str_to_bytes"
]



from .Decorators import overload
import json


class Serializable:
    IDENTIFIER_LABEL = "__IDENTIFIER__"

    def __init__(self):
        pass

    def to_json(self):
        return self.__dict__

    def serialize(self, path: str, JSON: bool = True) -> str:
        d = self.__dict__
        d[Serializable.IDENTIFIER_LABEL] = f"{type(self).__module__}.{type(self).__qualname__}"
        with open(path, "w") as file:
            file.write(json.dumps(
                self, default=lambda _: d, indent=4))

    @staticmethod
    def __static_deserialize(path: str, globals, JSON: bool = True):
        with open(path, "r") as file:
            d = json.loads(file.read())
            type_name = d[Serializable.IDENTIFIER_LABEL]
            classes = [v for v in globals]
            return d

    @classmethod
    def deserialize(cls, path: str, globals=None, JSON: bool = True):
        if cls == Serializable:
            return Serializable.__static_deserialize(path, globals, JSON)
        return Serializable.__class_deserialize(cls, path, JSON)

    def __class_deserialize(cls, path: str, JSON: bool = True):
        with open(path, "r") as file:
            d = json.loads(file.read())
            # res = cls
            # res.__dict__ = d
            return d


__all__ = [
    # "Serializable"
]



from ..Typing import Union
from ..Decorators import validate


@validate([int, float])
def sign(v: Union[int, float]) -> int:
    if v == 0:
        return 1
    return abs(v)/v


__all__ = [
    "sign"
]



from .Typing import Callable, Any, Union
import functools
import threading
from .Functions import areoneof, isoneof, isoneof_strict, isoftype
from .Exceptions import OverloadDuplication, OverloadNotFound, ValidationTypeError, ValidationValueError, ValidationReturnTypeError, TimeoutError
__validation_set = set()
__validation_instantiation_rule = dict()


def __validate_type(func: Callable, v: Any, T: type, validation_func: Callable[[Any], bool] = isoftype, msg: str = None) -> None:
    if not validation_func(v, T):
        raise ValidationTypeError(
            msg or f"In {func.__module__}.{func.__qualname__}(...)\nThe argument is: '{ v.__qualname__ if hasattr(v, '__qualname__') else v}'\nIt has the type of '{type(v)}'\nIt is marked as type(s): '{T}'")


def __validate_condition(func: Callable, v: Any, constraint: Callable[[Any], bool], msg: str = None) -> None:
    if not constraint(v):
        raise ValidationValueError(
            msg or f"In {func.__module__}.{func.__qualname__}(...)\nThe argument '{str(v)}' has failed provided constraint\nConstraint in {constraint.__module__}.{constraint.__qualname__}")


def __validate_arg(func: Callable, curr_arg: Any, curr_inner_arg: Any) -> None:
    if isoneof(curr_arg, [list, tuple]):
        # multiple type only:
        if areoneof(curr_arg, [type]):
            __validate_type(func, curr_inner_arg, curr_arg, isoneof)

        else:  # maybe with condition:
            class_type, constraint = curr_arg[0], curr_arg[1]

            # Type validation
            if isoneof(class_type, [list, tuple]):
                __validate_type(func, curr_inner_arg, class_type, isoneof)
            else:
                __validate_type(func, curr_inner_arg, class_type, isinstance)

            # constraints validation
            if constraint is not None:
                message = curr_arg[2] if len(curr_arg) > 2 else None
                __validate_condition(func, curr_inner_arg, constraint, message)
    else:
        __validate_type(func, curr_inner_arg, curr_arg)


def validate(*args, return_type=None, can_instantiate_multiple_times: bool = False) -> Callable:
    """validate decorator

        Is passed types of variables to perform type checking over\n
        The arguments must be passed in the same order\n

    for each parameter respectively you can choose one of four options:\n
        1. None - to skip\n
        2. Type - a type to check \n
        3. Sequence of Type to check if the type is contained in the sequence\n
        4. Sequence that contains three arguments:\n
            4.1 a Type or Sequence[Type]\n
            4.2 a function to call on argument\n
            4.3 a str to display in a ValueError iff the condition from 4.2 fails\n
    In addition you can use keyword 'return_type' for the returned value same as specified in 1,2,3
    """
    from .Exceptions import ValidationDuplicationError, ValidationTypeError, ValidationValueError, ValidationReturnTypeError

    def deco(func: Callable) -> Callable:
        if not isinstance(func, Callable):
            raise ValueError("validate decorator must decorate a callable")
        global __validation_set, __validation_instantiation_rule
        func_id = f"{func.__module__}.{func.__qualname__}"

        if func_id not in __validation_instantiation_rule:
            __validation_instantiation_rule[func_id] = can_instantiate_multiple_times
        assert can_instantiate_multiple_times == __validation_instantiation_rule[
            func_id], "can't change instantiation status on runtime"

        if func_id not in __validation_set:
            __validation_set.add(func_id)
        else:
            if not __validation_instantiation_rule[func_id]:
                raise ValidationDuplicationError(
                    "validate decorator is being used on two functions in the same module with the same name\nmaybe use @overload instead")

        @ functools.wraps(func)
        def wrapper(*inner_args, **inner_kwargs) -> Any:
            for i in range(min(len(args), len(inner_args))):
                if args[i] is not None:
                    __validate_arg(func, args[i], inner_args[i])
            res = func(*inner_args, **inner_kwargs)
            if return_type is not None:
                msg = f"In {func.__module__}.{func.__qualname__}(...)\nThe returned value is: '{ res.__qualname__ if hasattr(res, '__qualname__') else res}'\nIt has the type of '{type(res)}'\nIt is marked as type(s): '{return_type}'"
                __validate_type(func, res, return_type, msg=msg)
            return res
        return wrapper
    return deco


@ validate(Callable)
def NotImplemented(func: Callable) -> Callable:
    """decorator to mark function as not implemented for development purposes

    Args:
        func (Callable): the function to decorate
    """
    @ functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        raise NotImplementedError(
            f"As marked by the developer {func.__module__}.{func.__qualname__} is not implemented yet..")
    return wrapper


@ validate(Callable)
def PartiallyImplemented(func: Callable) -> Callable:
    """decorator to mark function as not fully implemented for development purposes

    Args:
        func (Callable): the function to decorate
    """
    from .Colors import warning

    @ functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        warning(
            f"As marked by the developer, {func.__module__}.{func.__qualname__} may not be fully implemented and might not work properly")
        return func(*args, **kwargs)
    return wrapper


@ validate(Callable)
def memo(func: Callable) -> Callable:
    """decorator to memorize function calls in order to improve performance by using more memory

    Args:
        func (Callable): function to memorize
    """
    cache: dict[tuple, Any] = {}

    @ functools.wraps(func)
    def wrapper(*args, **kwargs):
        if (args, *kwargs.items()) not in cache:
            cache[(args, *kwargs.items())] = func(*args, **kwargs)
        return cache[(args, *kwargs.items())]
    return wrapper


__overload_dict: dict[str, dict[tuple, Callable]] = dict()


def overload(*types) -> Callable:
    """decorator for overloading functions\n
    Usage\n-------\n
    @overload(str,str)\n
    def print_info(name,color):
        ...\n\n
    @overload(str,[int,float]))\n
    def print_info(name,age):
        ...\n\n

    * use None to skip argument
    * use no arguments to mark as default function
    * you should overload in decreasing order of specificity! e.g @overload(int) should appear in the code before @overload(Any)

    \n\n\n
    \nRaises:
        OverloadDuplication: if a functions is overloaded twice (or more) with same argument types
        OverloadNotFound: if an overloaded function is called with types that has no variant of the function

    \nNotice:
        The function's __doc__ will hold the value of the last variant only
    """
    # make sure to use unique global dictionary
    if len(types) == 1 and type(types[0]).__name__ == "function":
        raise ValueError("can't create an overload without defining types")
    global __overload_dict
    # allow input of both tuples and lists for flexibly
    if len(types) > 0:
        types = list(types)
        for i, maybe_list_of_types in enumerate(types):
            if isoneof(maybe_list_of_types, [list, tuple]):
                types[i] = tuple(sorted(list(maybe_list_of_types),
                                        key=lambda sub_type: sub_type.__name__))
        types = tuple(types)

    def deco(func: Callable) -> Callable:
        if not isinstance(func, Callable):
            raise TypeError("overload decorator must be used on a callable")

        # assign current overload to overload dictionary
        name = f"{func.__module__}.{func.__qualname__}"

        if name not in __overload_dict:
            __overload_dict[name] = dict()

        if types in __overload_dict[name]:
            # raise if current overload already exists for current function
            raise OverloadDuplication(
                f"{name} has duplicate overloading for type(s): {types}")

        __overload_dict[name][types] = func

        @ functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            default_func = None
            # select correct overload
            for variable_types, curr_func in __overload_dict[f"{func.__module__}.{func.__qualname__}"].items():
                if len(variable_types) == 0:
                    if default_func is None:
                        default_func = curr_func
                        continue
                    else:
                        # will not reach here because of duplicate overloading so this is redundant
                        raise ValueError("Can't have two default functions")

                if len(variable_types) != len(args):
                    continue

                for i, variable_type in enumerate(variable_types):
                    if variable_type is not None:
                        if isoneof(variable_type, [list, tuple]):
                            if not isoneof_strict(args[i], variable_type):
                                break
                        else:
                            if not isoftype(args[i], variable_type):
                                break
                else:
                    return curr_func(*args, **kwargs)

            if default_func is not None:
                return default_func(*args, **kwargs)
            # or raise exception if no overload exists for current arguments
            raise OverloadNotFound(
                f"function {func.__module__}.{func.__qualname__} is not overloaded with {[type(v) for v in args]}")

        return wrapper
    return deco


@ validate(Callable)
def abstractmethod(func: Callable) -> Callable:
    """A decorator to mark a function to be 'pure virtual' / 'abstract'

    Args:
        func (Callable): the function to mark

    Raises:
        NotImplementedError: the error that will rise when the marked function will be called if not overridden in a derived class
    """
    @ functools.wraps(func)
    def wrapper(*args, **kwargs):
        raise NotImplementedError(
            f"{func.__module__}.{func.__qualname__} MUST be overrided in a child class")
    return wrapper


purevirtual = abstractmethod

# __virtualization_tables = dict()


# @NotImplemented
# def virtual(func: Callable) -> Callable:
#     def wrapper(*args, **kwargs):
#         return func(*args, **kwargs)
#     return wrapper


# @NotImplemented
# def override(func: Callable) -> Callable:
#     def wrapper(*args, **kwargs):
#         return func(*args, **kwargs)
#     return wrapper


@ PartiallyImplemented
@ validate([str, Callable])
def deprecate(obj: Union[str, Callable] = None) -> Callable:
    """decorator to mark function as deprecated

    Args:
        obj (Union[str, None, Callable], optional): Defaults to None.

        Can operate in two configurations:\n
        1. obj is the function that you want to deprecate\n
        \t@deprecate\n
        \tdef foo(...):\n
        \t\t...\n\n
        2. obj is an advise message\n
        \t@deprecate("instead use ...")\n
        \tdef foo(...):
        \t\t...
    """
    from .Colors import warning
    # if callable(obj):
    if isinstance(obj, Callable):
        @ functools.wraps(obj)
        def wrapper(*args, **kwargs) -> Any:
            warning(
                f"As marked by the developer, {obj.__module__}.{obj.__qualname__} is deprecated")
            return obj(*args, **kwargs)
        return wrapper

    def deco(func: Callable) -> Callable:
        @ functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            warning(
                f"As marked by the developer, {func.__module__}.{func.__qualname__} is deprecated")
            if obj:
                print(obj)
            return func(*args, **kwargs)
        return wrapper
    return deco


@ validate(Callable)
def atomic(func):
    import threading
    lock = threading.Lock()

    @ functools.wraps(func)
    def wrapper(*args, **kwargs):
        with lock:
            return func(*args, **kwargs)
    return wrapper


@ validate([int, lambda d: d > 0, "limit_recursion's max_depth must be a positive integer"], None, bool)
def limit_recursion(max_depth: int, return_value=None, quiet: bool = True):
    """decorator to limit recursion of functions

    Args:
        max_depth (int): max recursion depth which is allowed for this function
        return_value (_type_, optional): The value to return when the limit is reached. Defaults to None.
            if is None, will return the last (args, kwargs)
        quiet (bool, optional): whether to print a warning message. Defaults to True.
    """
    import traceback
    import re
    from .Colors import warning

    def deco(func):
        @ functools.wraps(func)
        def wrapper(*args, **kwargs):
            depth = functools.reduce(
                lambda count, line:
                    count + 1 if re.search(f"{func.__name__}\(.*\)$", line)
                    else count,
                traceback.format_stack(), 0
            )
            if depth >= max_depth:
                if not quiet:
                    warning(
                        f"limit_recursion has limited the number of calls for {func.__module__}.{func.__qualname__} to {max_depth}")
                if return_value:
                    return return_value
                return args, kwargs
            return func(*args, **kwargs)
        return wrapper
    return deco


@validate([int, float])
def timeout(timeout: int | float) -> Callable:
    """A decorator to limit runtime for a function

    Args:
        timeout (int | float): allowed runtime duration

    Raises:
        thread_error: if there is a thread related error
        function_error: if there is an error in the decorated function

    Returns:
        Callable: The decorated function
    """
    # https://stackoverflow.com/a/21861599/6416556
    def timeout_deco(func: Callable) -> Callable:
        if not isinstance(func, Callable):
            raise ValueError("timeout must decorate a function")

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            res = [
                TimeoutError(f'{func.__module__}.{func.__qualname__} timed out after {timeout} seconds!')]

            def timeout_wrapper() -> None:
                try:
                    res[0] = func(*args, **kwargs)
                except Exception as function_error:
                    res[0] = function_error

            t = threading.Thread(target=timeout_wrapper, daemon=True)
            try:
                t.start()
                t.join(timeout)
            except Exception as thread_error:
                raise thread_error
            if isinstance(res[0], BaseException):
                raise res[0]
            return res[0]
        return wrapper
    return timeout_deco


@validate(Callable, Callable)
def attach(before: Callable = None, after: Callable = None) -> Callable:
    if before is None and after is None:
        raise ValueError("You must supply at least one function")

    def attach_deco(func: Callable):
        if not isinstance(func, Callable):
            raise ValueError("attach must decorate a function")

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if before is not None:
                before()
            res = func(*args, **kwargs)
            if after is not None:
                after()
            return res
        return wrapper
    return attach_deco


__all__ = [
    "validate",
    "NotImplemented",
    "PartiallyImplemented",
    "memo",
    "overload",
    "abstractmethod",
    "purevirtual",
    # "virtual",
    # "override",
    "deprecate",
    "atomic",
    "limit_recursion",
    "timeout",
    "attach"
]
